# -*- coding: utf-8 -*-
import logging

from .utils import make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
_safety_fixture, _safety_state, run_cell_ = make_safety_fixture()


def run_cell(cell, cell_id=None):
    """Mocks the `change active cell` portion of the comm protocol"""
    if cell_id is not None:
        _safety_state[0].handle({
            'type': 'change_active_cell',
            'active_cell_id': cell_id
        })
    run_cell_(cell)


def list_to_dict(lst):
    return dict((i, val) for i, val in enumerate(lst))


def test_simple():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'x = 42',
        3: 'logging.info(y)',
    }
    run_cell(cells[0])
    run_cell(cells[1])
    run_cell(cells[2])
    response = _safety_state[0].multicell_precheck(cells)
    assert response['stale_input_cells'] == [3]
    assert response['stale_output_cells'] == []
    assert response['stale_links'] == {3: [1]}
    assert response['refresher_links'] == {1: [3]}


def test_refresh_after_exception_fixed():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'logging.info(y)',
    }
    run_cell(cells[0], 0)
    run_cell(cells[2], 2)
    run_cell(cells[1], 1)
    response = _safety_state[0].multicell_precheck(cells)
    assert response['stale_output_cells'] == [2]


def test_refresh_after_val_changed():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'logging.info(y)',
        3: 'y = 42',
    }
    run_cell(cells[0], 0)
    run_cell(cells[1], 1)
    run_cell(cells[2], 2)
    run_cell(cells[3], 3)
    response = _safety_state[0].multicell_precheck(cells)
    assert response['stale_output_cells'] == [2]



# -*- coding: utf-8 -*-
import logging
from typing import cast, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Union

logger = logging.getLogger(__name__)


def retrieve_namespace_attr_or_sub(obj: 'Any', attr_or_sub: 'Union[str, int]', is_subscript: bool):
    try:
        if is_subscript:
            # TODO: more complete list of things that are checkable
            #  or could cause side effects upon subscripting
            if isinstance(obj, dict) and attr_or_sub not in obj:
                raise KeyError()
            else:
                return obj[attr_or_sub]
        else:
            assert isinstance(attr_or_sub, str)
            if not hasattr(obj, attr_or_sub):
                raise AttributeError()
            else:
                return getattr(obj, attr_or_sub)
    except (KeyError, IndexError, AttributeError):
        raise
    # except AssertionError as e:
    #     print(obj, attr_or_sub, is_subscript)
    #     raise e
    except Exception as e:
        logger.warning('unexpected exception: %s' % e)
        raise e



# -*- coding: utf-8 -*-
import ast

from nbsafety.analysis.lineno_stmt_map import compute_lineno_to_stmt_mapping


def test_for_loop():
    code = """
for i in range(10):
    a: int = i
    b = a + i
    lst: List[int] = [a, b]
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    assert isinstance(mapping[1], ast.For)
    assert isinstance(mapping[2], ast.AnnAssign)
    assert isinstance(mapping[3], ast.Assign)
    assert isinstance(mapping[4], ast.AnnAssign)


def test_multiline_for_loop():
    code = """
for i in [
    0,
    1,
    2,
    3,
    4,
]:
    a = i
    b = a + i
    lst = [a, b]
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    for i in range(1, 7):
        assert isinstance(mapping[i], ast.For)
    assert 7 not in mapping
    assert isinstance(mapping[8], ast.Assign)
    assert isinstance(mapping[9], ast.Assign)
    assert isinstance(mapping[10], ast.Assign)



# -*- coding: utf-8 -*-
import ast
from collections import defaultdict
from contextlib import contextmanager
import logging
import sys
from typing import TYPE_CHECKING

from IPython import get_ipython
from IPython.core.magic import register_cell_magic, register_line_magic

from .analysis import AttrSubSymbolChain, ComputeLiveSymbolRefs, compute_lineno_to_stmt_mapping
from .ipython_utils import (
    ast_transformer_context,
    cell_counter,
    run_cell,
    save_number_of_currently_executing_cell,
)
from . import line_magics
from .scope import Scope, NamespaceScope
from .tracing import AttrSubTracingManager, make_tracer, TraceState

if TYPE_CHECKING:
    from typing import Any, Dict, List, Set, Optional, Tuple, Union
    from .analysis import SymbolRef
    from .data_symbol import DataSymbol

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

_MAX_WARNINGS = 10
_SAFETY_LINE_MAGIC = 'safety'


def _safety_warning(node: 'DataSymbol'):
    if node.has_stale_ancestor:
        required_cell_num = node.required_cell_num
        fresher_ancestors = node.fresher_ancestors
    else:
        raise ValueError('Expected node with stale ancestor; got %s' % node)
    logger.warning(
        f'`{node.readable_name}` defined in cell {node.defined_cell_num} may depend on '
        f'old version(s) of [{", ".join(f"`{str(dep)}`" for dep in fresher_ancestors)}] '
        f'(latest update in cell {required_cell_num}).'
    )


class DependencySafety(object):
    """Holds all the state necessary to detect stale dependencies in Jupyter notebooks."""
    def __init__(self, cell_magic_name=None, use_comm=False, **kwargs):
        # Note: explicitly adding the types helps PyCharm's built-in code inspection
        self.namespaces: Dict[int, NamespaceScope] = {}
        self.aliases: Dict[int, Set[DataSymbol]] = defaultdict(set)
        self.global_scope: Scope = Scope(self)
        self.updated_symbols: Set[DataSymbol] = set()
        self.updated_scopes: Set[NamespaceScope] = set()
        self.garbage_namespace_obj_ids: Set[int] = set()
        self.statement_cache: Dict[int, Dict[int, ast.stmt]] = {}
        self.trace_event_counter: List[int] = [0]
        self.stale_dependency_detected = False
        self.trace_state: TraceState = TraceState(self)
        self.attr_trace_manager: AttrSubTracingManager = AttrSubTracingManager(self, self.global_scope, self.trace_event_counter)
        self.store_history: bool = kwargs.pop('store_history', True)
        self.use_comm: bool = use_comm
        self.trace_messages_enabled: bool = kwargs.pop('trace_messages_enabled', False)
        self._last_execution_counter = 0
        self._counters_by_cell_id: Dict[Union[str, int], int] = {}
        self._active_cell_id: Optional[str] = None
        self._save_prev_trace_state_for_tests: bool = kwargs.pop('save_prev_trace_state_for_tests', False)
        if self._save_prev_trace_state_for_tests:
            self.prev_trace_state: Optional[TraceState] = None
        if cell_magic_name is None:
            self._cell_magic = None
        else:
            self._cell_magic = self._make_cell_magic(cell_magic_name)
        self._line_magic = self._make_line_magic()
        self._last_refused_code: Optional[str] = None
        self.no_stale_propagation_for_same_cell_definition = True
        self._track_dependencies = True

        self._disable_level = 0
        self._prev_cell_nodes_with_stale_deps: Set[DataSymbol] = set()

        if self.use_comm:
            get_ipython().kernel.comm_manager.register_target(__package__, self._comm_target)

    def _comm_target(self, comm, open_msg):
        @comm.on_msg
        def _responder(msg):
            request = msg['content']['data']
            self.handle(request, comm=comm)

        comm.send({'type': 'establish'})

    def handle(self, request, comm=None):
        if request['type'] == 'change_active_cell':
            cell_id = request['active_cell_id']
            self._active_cell_id = cell_id
        elif request['type'] == 'cell_freshness':
            cell_id = request['executed_cell_id']
            self._counters_by_cell_id[cell_id] = self._last_execution_counter
            cells_by_id = request['content_by_cell_id']
            response = self.multicell_precheck(cells_by_id)
            response['type'] = 'cell_freshness'
            if comm is not None:
                comm.send(response)
        else:
            logger.error('Unsupported request type for request %s' % request)

    def multicell_precheck(self, cells_by_id: 'Dict[Union[int, str], str]') -> 'Dict[str, Any]':
        stale_input_cells = []
        stale_output_cells = []
        fresh_cells = []
        for cell_id, cell_content in cells_by_id.items():
            try:
                stale_nodes, max_defined_cell_num = self._precheck_stale_nodes(cell_content)
                if len(stale_nodes) > 0:
                    stale_input_cells.append(cell_id)
                elif max_defined_cell_num > self._counters_by_cell_id.get(cell_id, float('inf')):
                    stale_output_cells.append(cell_id)
                else:
                    fresh_cells.append(cell_id)
            except SyntaxError:
                continue
        stale_links = defaultdict(list)
        refresher_links = defaultdict(list)
        for candidate_refresher_cell_id in fresh_cells + stale_output_cells:
            candidate_refresher_cell = cells_by_id[candidate_refresher_cell_id]
            for stale_cell_id in stale_input_cells:
                try:
                    if not self._precheck_simple(f'{candidate_refresher_cell}\n{cells_by_id[stale_cell_id]}'):
                        stale_links[stale_cell_id].append(candidate_refresher_cell_id)
                        refresher_links[candidate_refresher_cell_id].append(stale_cell_id)
                except SyntaxError:
                    continue
        return {
            'stale_input_cells': stale_input_cells,
            'stale_output_cells': stale_output_cells,
            'stale_links': stale_links,
            'refresher_links': refresher_links,
        }

    @staticmethod
    def _get_cell_ast(cell):
        lines = []
        for line in cell.strip().split('\n'):
            # TODO: figure out more robust strategy for filtering / transforming lines for the ast parser
            # normally filter line magic, but for %time, try to actually trace the statement being timed
            # TODO: how to do this?
            if line.startswith('%'):
                if line.startswith('%time '):
                    lines.append(line[len('%time '):].strip())
                else:
                    continue
            elif line.startswith('!'):
                continue
            elif line.startswith('cd'):
                continue
            elif line.endswith('?'):
                continue
            lines.append(line)
        return ast.parse('\n'.join(lines))

    def compute_live_symbol_refs(self, code: 'Union[ast.Module, str]') -> 'Set[SymbolRef]':
        if isinstance(code, str):
            code = ast.parse(code)
        return ComputeLiveSymbolRefs(self)(code)

    def _precheck_stale_nodes(self, cell: 'Union[ast.Module, str]'):
        if isinstance(cell, str):
            cell = self._get_cell_ast(cell)
        stale_nodes = set()
        max_defined_cell_num = -1
        for symbol_ref in self.compute_live_symbol_refs(cell):
            if isinstance(symbol_ref, str):
                node = self.global_scope.lookup_data_symbol_by_name_this_indentation(symbol_ref)
            elif isinstance(symbol_ref, AttrSubSymbolChain):
                node = self.global_scope.get_most_specific_data_symbol_for_attrsub_chain(symbol_ref, self.namespaces)
            else:
                logger.warning('invalid type for ref %s', symbol_ref)
                continue
            if node is None:
                continue
            # print(node, node.has_stale_ancestor)
            max_defined_cell_num = max(max_defined_cell_num, node.defined_cell_num)
            if node.has_stale_ancestor:
                stale_nodes.add(node)
            if node.obj_id in self.namespaces:
                namespace_scope = self.namespaces[node.obj_id]
                max_defined_cell_num = max(max_defined_cell_num, namespace_scope.max_defined_timestamp)
        return stale_nodes, max_defined_cell_num

    def _precheck_simple(self, cell):
        return len(self._precheck_stale_nodes(cell)[0]) > 0

    def _precheck_for_stale(self, cell: str):
        # Precheck process. First obtain the names that need to be checked. Then we check if their
        # `defined_cell_num` is greater than or equal to required; if not we give a warning and return `True`.
        try:
            cell_ast = self._get_cell_ast(cell)
        except SyntaxError:
            return False
        self.statement_cache[cell_counter()] = compute_lineno_to_stmt_mapping(cell_ast)
        if self._last_refused_code is None or cell != self._last_refused_code:
            self._prev_cell_nodes_with_stale_deps, _ = self._precheck_stale_nodes(cell_ast)
            if len(self._prev_cell_nodes_with_stale_deps) > 0 and self._disable_level < 2:
                warning_counter = 0
                for node in self._prev_cell_nodes_with_stale_deps:
                    if warning_counter >= _MAX_WARNINGS:
                        logger.warning(f'{len(self._prev_cell_nodes_with_stale_deps) - warning_counter}'
                                       ' more nodes with stale dependencies skipped...')
                        break
                    _safety_warning(node)
                    warning_counter += 1
                self.stale_dependency_detected = True
                self._last_refused_code = cell
                if self._disable_level == 0:
                    return True
        else:
            # Instead of breaking the dependency chain, simply refresh the nodes
            # with stale deps to their required cell numbers
            for node in self._prev_cell_nodes_with_stale_deps:
                node.defined_cell_num = node.required_cell_num
                node.namespace_data_syms_with_stale = set()
                node.fresher_ancestors = set()
            self._prev_cell_nodes_with_stale_deps.clear()

        self._last_refused_code = None
        return False

    def safe_execute(self, cell: str, run_cell_func):
        if self._disable_level == 3:
            return run_cell_func(cell)

        with save_number_of_currently_executing_cell():
            self._last_execution_counter = cell_counter()
            if self._active_cell_id is not None:
                self._counters_by_cell_id[self._active_cell_id] = self._last_execution_counter
                self._active_cell_id = None
            # Stage 1: Precheck.
            if self._precheck_for_stale(cell):
                # FIXME: hack to increase cell number
                #  ideally we shouldn't show a cell number at all if we fail precheck since nothing executed
                return run_cell_func('None')

            def _backup():
                # something went wrong silently (e.g. due to line magic); fall back to just executing the code
                logger.warning('Something failed while attempting traced execution; '
                               'falling back to uninstrumented execution.')
                return run_cell_func(cell)

            # Stage 2: Trace / run the cell, updating dependencies as they are encountered.
            try:
                with self._tracing_context():
                    ret = run_cell_func(cell)
            finally:
                if self.trace_state.error_occurred:
                    ret = _backup()
                return ret

    def _make_cell_magic(self, cell_magic_name):
        def _run_cell_func(cell):
            run_cell(cell, store_history=self.store_history)

        def _dependency_safety(_, cell: str):
            self.safe_execute(cell, _run_cell_func)

        # FIXME (smacke): probably not a great idea to rely on this
        _dependency_safety.__name__ = cell_magic_name
        return register_cell_magic(_dependency_safety)

    @contextmanager
    def _tracing_context(self):
        sys.settrace(make_tracer(self))
        try:
            with ast_transformer_context(self.attr_trace_manager.ast_transformer):
                yield
        finally:
            sys.settrace(None)
            # TODO: actually handle errors that occurred in our code while tracing
            if not self.trace_state.error_occurred:
                self._reset_trace_state_hook()
            for updated_symbol in self.updated_symbols:
                updated_symbol.refresh()
            for updated_scope in self.updated_scopes:
                updated_scope.refresh()
            self.updated_symbols.clear()
            self.updated_scopes.clear()

    def _reset_trace_state_hook(self):
        if self.dependency_tracking_enabled and self.trace_state.prev_trace_stmt_in_cur_frame is not None:
            self.trace_state.prev_trace_stmt_in_cur_frame.finished_execution_hook()
        self.attr_trace_manager.reset()  # should happen on finish_execution_hook, but since its idempotent do it again
        if self._save_prev_trace_state_for_tests:
            self.prev_trace_state = self.trace_state
        self.trace_state = TraceState(self)
        self._gc()

    def _make_line_magic(self):
        def _safety(line_: str):
            line = line_.split()
            if not line or line[0] not in [
                "show_graph", "show_deps", "show_dependency", "show_dependencies",
                "show_stale", "set_disable_level", "set_propagation",
                "trace_messages",
                "remove_dependency", "add_dependency", "turn_off_warnings_for", "turn_on_warnings_for",
            ]:
                print(line_magics.USAGE)
                return
            if line[0] == "show_graph":
                return line_magics.show_graph(self)
            elif line[0] in ("show_deps", "show_dependency", "show_dependencies"):
                return line_magics.show_deps(self, line)
            elif line[0] == "show_stale":
                return line_magics.show_stale(self, line)
            elif line[0] == "set_disable_level":
                return line_magics.set_disable_level(self, line)
            elif line[0] == "set_propagation":
                return line_magics.set_propagation(self, line)
            elif line[0] == "trace_messages":
                return line_magics.configure_trace_messages(self, line)
            elif line[0] == "remove_dependency":
                return line_magics.remove_dep(self, line)
            elif line[0] == "add_dependency":
                return line_magics.add_dep(self, line)
            elif line[0] == "turn_off_warnings_for":
                return line_magics.turn_off_warnings_for(self, line)
            elif line[0] == "turn_on_warnings_for":
                return line_magics.turn_on_warnings_for(self, line)

        # FIXME (smacke): probably not a great idea to rely on this
        _safety.__name__ = _SAFETY_LINE_MAGIC
        return register_line_magic(_safety)

    @property
    def dependency_tracking_enabled(self):
        return self._track_dependencies

    @property
    def cell_magic_name(self):
        return self._cell_magic.__name__

    @property
    def line_magic_name(self):
        return self._line_magic.__name__

    def all_data_symbols(self):
        for alias_set in self.aliases.values():
            for alias in alias_set:
                yield alias

    def test_and_clear_detected_flag(self):
        ret = self.stale_dependency_detected
        self.stale_dependency_detected = False
        return ret

    def namespace_gc(self):
        for obj_id in self.garbage_namespace_obj_ids:
            self.namespaces.pop(obj_id, None)
        self.garbage_namespace_obj_ids.clear()

    def _gc(self):
        for dsym in list(self.all_data_symbols()):
            if dsym.is_garbage:
                dsym.collect_self_garbage()



# -*- coding: utf-8 -*-
import ast
import logging
from typing import TYPE_CHECKING

from .attr_symbols import get_attrsub_symbol_chain, AttrSubSymbolChain
from .mixins import SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin

if TYPE_CHECKING:
    from typing import List, Set, Union
    from ..safety import DependencySafety
    from ..types import SymbolRef

logger = logging.getLogger(__name__)


# TODO: have the logger warnings additionally raise exceptions for tests
class ComputeLiveSymbolRefs(ast.NodeVisitor):
    def __init__(self, safety: 'DependencySafety'):
        self.safety = safety
        self.killed: Set[Union[str, AttrSubSymbolChain]] = set()

    def __call__(self, module_node: ast.Module):
        """
        This function should be called when we want to precheck an ast.Module. For
        each line/block of the cell we first run the check of new assignments, then
        we obtain all the names. In these names, we put the ones that are user
        defined and not in the safe_set into the return check_set for further
        checks.
        """
        check_set = set()
        for node in module_node.body:
            self.visit(node)
            for ref in _get_all_symbol_refs(node):
                if ref in self.killed:
                    continue
                # TODO: check for all subchains in the safe set, not just the first symbol
                if isinstance(ref, AttrSubSymbolChain):
                    leading_symbol = ref.symbols[0]
                    if isinstance(leading_symbol, str) and leading_symbol in self.killed:
                        continue
                check_set.add(ref)
        # print(self.safe_set)
        # print(check_set)
        return check_set

    # In case of assignment, we put the new assigned variable into a safe_set
    # to indicate that we know for sure it won't have stale dependency.  Note
    # that node.targets might contain multiple ast.Name node in the case of
    # "a = b = 3", so we go through each node in the targets.  Also note that
    # `target` would be an ast.Tuple node in the case of "a,b = 3,4". Thus
    # we need to break the tuple in that case.
    def visit_Assign(self, node: ast.Assign):
        for target_node in node.targets:
            if isinstance(target_node, ast.Tuple):
                for element_node in target_node.elts:
                    if isinstance(element_node, ast.Name):
                        self.killed.add(element_node.id)
            else:
                self.visit_Assign_or_AugAssign_target(target_node)

    def visit_AnnAssign(self, node: ast.AnnAssign):
        self.visit_Assign_or_AugAssign_target(node.target)

    def visit_AugAssign(self, node: ast.AugAssign):
        self.visit_Assign_or_AugAssign_target(node.target)

    def visit_Assign_or_AugAssign_target(self, target_node: 'Union[ast.Attribute, ast.Name, ast.Subscript, ast.expr]'):
        if isinstance(target_node, ast.Name):
            self.killed.add(target_node.id)
        elif isinstance(target_node, (ast.Attribute, ast.Subscript)):
            self.killed.add(get_attrsub_symbol_chain(target_node))
        else:
            logger.warning('unsupported type for node %s' % target_node)

    # We also put the name of new functions in the safe_set
    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.killed.add(node.name)

    def visit_For(self, node: ast.For):
        # Case "for a,b in something: "
        if isinstance(node.target, ast.Tuple):
            for name_node in node.target.elts:
                if isinstance(name_node, ast.Name):
                    self.killed.add(name_node.id)
                else:
                    logger.warning('unsupported type for node %s' % name_node)
        # case "for a in something"
        elif isinstance(node.target, ast.Name):
            self.killed.add(node.target.id)
        else:
            logger.warning('unsupported type for node %s' % node.target)

        # Then we keep doing the visit for the body of the loop.
        for line in node.body:
            self.visit(line)


# Call GetAllNames()(ast_tree) to get a set of all names appeared in ast_tree.
# Helper Class
class GetAllSymbolRefs(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self):
        self.ref_set: Set[SymbolRef] = set()
        self.inside_attrsub = False
        self.skip_simple_names = False

    def __call__(self, node: ast.AST):
        self.visit(node)
        return self.ref_set

    def attrsub_context(self, inside=True):
        return self.push_attributes(inside_attrsub=inside, skip_simple_names=inside)

    def args_context(self):
        return self.push_attributes(skip_simple_names=False)

    def visit_Name(self, node: ast.Name):
        if not self.skip_simple_names:
            self.ref_set.add(node.id)

    # We overwrite FunctionDef because we don't need to check names in the body of the definition.
    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.visit(node.args)

    def visit_ClassDef(self, node: ast.ClassDef):
        self.generic_visit(node.bases)
        self.generic_visit(node.decorator_list)

    def visit_Call(self, node: ast.Call):
        with self.args_context():
            self.generic_visit(node.args)
            for kwarg in node.keywords:
                self.visit(kwarg.value)
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            self.ref_set.add(get_attrsub_symbol_chain(node))
            with self.attrsub_context():
                self.visit(node.func)
        else:
            self.visit(node.func)

    def visit_Attribute(self, node: ast.Attribute):
        if not self.inside_attrsub:
            self.ref_set.add(get_attrsub_symbol_chain(node))
        with self.attrsub_context():
            self.visit(node.value)

    def visit_Subscript(self, node: ast.Subscript):
        if not self.inside_attrsub:
            self.ref_set.add(get_attrsub_symbol_chain(node))
        with self.attrsub_context():
            self.visit(node.value)
        with self.attrsub_context(inside=False):
            self.visit(node.slice)


def _get_all_symbol_refs(node: ast.AST):
    return GetAllSymbolRefs()(node)



# -*- coding: utf-8 -*-
import ast
import logging
from typing import cast, TYPE_CHECKING

from .trace_events import TraceEvent

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Tuple
    from types import FrameType
    from .trace_stmt import TraceStatement
    from ..safety import DependencySafety
    from ..scope import NamespaceScope

logger = logging.getLogger(__name__)


class TraceState(object):
    def __init__(self, safety: 'DependencySafety'):
        self.safety = safety
        self.cur_frame_scope = safety.global_scope
        self.prev_trace_stmt_in_cur_frame: Optional[TraceStatement] = None
        self.inside_lambda = False
        self.call_depth = 0
        self.traced_statements: Dict[int, TraceStatement] = {}
        self.stack: List[Tuple[TraceStatement, bool]] = []
        self.source: Optional[str] = None
        self.prev_trace_stmt: Optional[TraceStatement] = None
        self.prev_event: Optional[TraceEvent] = None
        self.error_occurred = False

    def _check_prev_stmt_done_executing_hook(self, event: 'TraceEvent', trace_stmt: 'TraceStatement'):
        if event not in (
                TraceEvent.line, TraceEvent.return_
        ) or self.prev_event in (
                TraceEvent.call, TraceEvent.exception
        ):
            return

        # we'll be needing these
        prev_this_frame = self.prev_trace_stmt_in_cur_frame
        prev_overall = self.prev_trace_stmt

        if prev_overall != trace_stmt:
            self.safety.attr_trace_manager.stmt_transition_hook()

        if event == TraceEvent.return_:
            if prev_overall is not None and prev_overall is not self.stack[-1][0]:
                prev_overall.finished_execution_hook()

        if self.prev_event == TraceEvent.return_:
            if prev_this_frame is not None:
                if len(self.stack) == 0 or prev_this_frame is not self.stack[-1][0]:
                    # this condition ensures we're not inside of a list comprehension or something with multiple calls
                    prev_this_frame.finished_execution_hook()
            return

        if prev_this_frame is None or prev_this_frame.marked_finished:
            return

        finished = prev_this_frame is not trace_stmt
        finished = finished and not (
            # classdefs are not finished until we reach the end of the class body
            isinstance(prev_this_frame.stmt_node, ast.ClassDef) and self.prev_event != TraceEvent.return_
        )
        if finished:
            prev_this_frame.finished_execution_hook()

    def state_transition_hook(
            self,
            event: 'TraceEvent',
            trace_stmt: 'TraceStatement'
    ):
        self.safety.trace_event_counter[0] += 1

        self._check_prev_stmt_done_executing_hook(event, trace_stmt)

        self.prev_trace_stmt = trace_stmt
        if event == TraceEvent.line:
            self.prev_trace_stmt_in_cur_frame = trace_stmt
        if event == TraceEvent.call:
            # TODO: figure out a better way to determine if we're inside a lambda
            #  could this one lead to a false negative if a lambda is in the default of a function def kwarg?
            inside_lambda = not isinstance(trace_stmt.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef))
            self.stack.append((self.prev_trace_stmt_in_cur_frame, self.inside_lambda))
            self.inside_lambda = inside_lambda
            # print('scope', trace_stmt.scope)
            if inside_lambda:
                self.cur_frame_scope = trace_stmt.get_post_call_scope(self.cur_frame_scope)
            else:
                with trace_stmt.replace_active_scope(self.safety.attr_trace_manager.active_scope_for_call):
                    # print('active scope', trace_stmt.scope)
                    self.cur_frame_scope = trace_stmt.get_post_call_scope(self.cur_frame_scope)
                    # print('post call scope', self.cur_frame_scope)
            logger.debug('entering scope %s', self.cur_frame_scope)
            self.prev_trace_stmt_in_cur_frame = None
            self.safety.attr_trace_manager.push_stack(self.cur_frame_scope)
        if event == TraceEvent.return_:
            logger.debug('leaving scope %s', self.cur_frame_scope)
            return_to_stmt, return_to_inside_lambda = self.stack.pop()
            assert return_to_stmt is not None
            if self.prev_event != TraceEvent.exception:
                # exception events are followed by return events until we hit an except clause
                # no need to track dependencies in this case
                if isinstance(return_to_stmt.stmt_node, ast.ClassDef):
                    return_to_stmt.class_scope = cast('NamespaceScope', self.cur_frame_scope)
                elif isinstance(trace_stmt.stmt_node, ast.Return) or self.inside_lambda:
                    return_to_stmt.call_point_deps.append(trace_stmt.compute_rval_dependencies())
            self.inside_lambda = return_to_inside_lambda
            # reset for the previous frame, so that we push it again if it has another funcall
            self.prev_trace_stmt_in_cur_frame = return_to_stmt
            # self.cur_frame_scope = return_to_stmt.scope
            self.safety.attr_trace_manager.pop_stack()
            self.cur_frame_scope = self.safety.attr_trace_manager.active_scope
            logger.debug('entering scope %s', self.cur_frame_scope)
        self.prev_event = event

    @staticmethod
    def get_position(frame: 'FrameType'):
        cell_num = int(frame.f_code.co_filename.split('-')[2])
        return cell_num, frame.f_lineno



# -*- coding: utf-8 -*-
import ast
from typing import cast, Union, TYPE_CHECKING

from ..utils.mixins import CommonEqualityMixin

if TYPE_CHECKING:
    from typing import List, Tuple, Union


class CallPoint(CommonEqualityMixin):
    def __init__(self, symbol: str):
        self.symbol = symbol

    def __hash__(self):
        return hash(self.symbol)

    def __repr__(self):
        return repr(str(self))

    def __str__(self):
        return self.symbol + '(...)'


class AttrSubSymbolChain(CommonEqualityMixin):
    def __init__(self, symbols: 'List[Union[str, CallPoint]]'):
        self.symbols: 'Tuple[Union[str, CallPoint], ...]' = tuple(symbols)
        self.call_points = tuple(filter(lambda x: isinstance(x, CallPoint), self.symbols))

    def __hash__(self):
        return hash(self.symbols)

    def __repr__(self):
        return repr(self.symbols)


class GetAttrSubSymbols(ast.NodeVisitor):
    def __init__(self):
        self.symbol_chain: List[Union[str, CallPoint]] = []

    def __call__(self, node: 'Union[ast.Attribute, ast.Subscript, ast.Call]') -> 'AttrSubSymbolChain':
        self.visit(node)
        self.symbol_chain.reverse()
        return AttrSubSymbolChain(self.symbol_chain)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute):
            self.symbol_chain.append(CallPoint(node.func.attr))
            self.visit(node.func.value)
        elif isinstance(node.func, ast.Subscript):
            self.symbol_chain.append(CallPoint(node.func.slice))
            self.visit(node.func.value)
        elif isinstance(node.func, ast.Name):
            self.symbol_chain.append(CallPoint(node.func.id))
        else:
            raise TypeError('invalid type for node.func %s' % node.func)

    def visit_Attribute(self, node):
        self.symbol_chain.append(node.attr)
        self.visit(node.value)

    def visit_Subscript(self, node):
        node_slice = node.slice
        if isinstance(node_slice, ast.Index):
            slice_index = node_slice.value
            if isinstance(slice_index, ast.Str):
                self.symbol_chain.append(slice_index.s)
            elif isinstance(slice_index, ast.Num):
                self.symbol_chain.append(slice_index.n)
            elif isinstance(slice_index, ast.Name):
                # FIXME: hack to make the static checker stop here
                # In the future, it should try to attempt to resolve
                # the value of the ast.Name node
                self.symbol_chain.append(CallPoint(slice_index.id))
            else:
                # give up
                pass
                # raise TypeError('unexpected type for node.slice %s' % node_slice)
        else:
            # give up
            pass
        self.visit(node.value)

    def visit_Name(self, node):
        self.symbol_chain.append(node.id)

    def visit_Str(self, node):
        return

    def generic_visit(self, node):
        # raise ValueError('we should never get here: %s' % node)
        # give up
        return


def get_attrsub_symbol_chain(maybe_node: Union[str, ast.Attribute, ast.Subscript, ast.Call]) -> AttrSubSymbolChain:
    if isinstance(maybe_node, (ast.Attribute, ast.Subscript, ast.Call)):
        node = maybe_node
    else:
        node = cast(Union[ast.Attribute, ast.Subscript, ast.Call], cast(ast.Expr, ast.parse(maybe_node).body[0]).value)
    if not isinstance(node, (ast.Attribute, ast.Subscript, ast.Call)):
        raise TypeError('invalid type for node %s' % node)
    return GetAttrSubSymbols()(node)



# -*- coding: utf-8 -*-
import ast
from typing import cast, TYPE_CHECKING

from .attr_symbols import get_attrsub_symbol_chain
from .mixins import SaveOffAttributesMixin, VisitListsMixin

if TYPE_CHECKING:
    from typing import Sequence, Union


class GetAssignmentLvalRvalSymbolRefs(SaveOffAttributesMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self):
        # TODO: figure out how to give these type annotations
        self.lval_symbols = []
        self.rval_symbols = []
        self.gather_rvals = True

    def __call__(self, node: ast.Assign):
        self.visit(node)

    def gather_lvals_context(self):
        return self.push_attributes(gather_rvals=False)

    def gather_rvals_context(self):
        return self.push_attributes(gather_rvals=True)

    @property
    def to_add_set(self):
        if self.gather_rvals:
            return self.rval_symbols
        else:
            return self.lval_symbols

    @to_add_set.setter
    def to_add_set(self, val):
        if self.gather_rvals:
            self.rval_symbols = val
        else:
            self.lval_symbols = val

    def visit_Name(self, node):
        self.to_add_set.append(node.id)

    def visit_Num(self, node):
        self.to_add_set.append(None)

    def visit_Str(self, node):
        self.to_add_set.append(None)

    def visit_NameConstant(self, node):
        self.to_add_set.append(None)

    def visit_Tuple(self, node):
        self.visit_List_or_Tuple(node)

    def visit_List(self, node):
        self.visit_List_or_Tuple(node)

    def visit_Dict(self, node):
        inner_symbols = []
        with self.push_attributes(to_add_set=inner_symbols):
            self.visit(node.keys)
            self.visit(node.values)
        self.to_add_set.append(inner_symbols)

    def visit_List_or_Tuple(self, node):
        inner_symbols = []
        with self.push_attributes(to_add_set=inner_symbols):
            self.visit(node.elts)
        self.to_add_set.append(inner_symbols)

    def visit_expr(self, node):
        assert self.gather_rvals
        inner_symbols = []
        with self.push_attributes(to_add_set=inner_symbols):
            # call super generic_visit since self generic_visit calls visit_expr
            super().generic_visit(node)
        self.to_add_set.append(inner_symbols)

    def generic_visit(self, node: 'Union[ast.AST, Sequence[ast.AST]]'):
        # The purpose of this is to make sure we call our visit_expr method if we see an expr
        if isinstance(node, ast.expr):
            self.visit_expr(node)
        else:
            super().generic_visit(node)

    def visit_Assign(self, node):
        with self.gather_lvals_context():
            for target in node.targets:
                target_lval_symbols = []
                with self.push_attributes(lval_symbols=target_lval_symbols):
                    self.visit(target)
                self.lval_symbols.append(target_lval_symbols)
        with self.gather_rvals_context():
            self.visit(node.value)

    def visit_Call(self, node):
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            self.to_add_set.append(get_attrsub_symbol_chain(node))
        else:
            self.generic_visit(node)

    def visit_Attribute_or_Subscript(self, node):
        # TODO: we'll ignore args inside of inner calls, e.g. f.g(x, y).h
        self.to_add_set.append(get_attrsub_symbol_chain(node))

    def visit_Attribute(self, node):
        self.visit_Attribute_or_Subscript(node)

    def visit_Subscript(self, node):
        self.visit_Attribute_or_Subscript(node)

    def visit_Keyword(self, node):
        self.visit(node.value)

    def visit_Starred(self, node):
        self.visit(node.value)

    def visit_Lambda(self, node):
        assert self.gather_rvals
        # remove node.arguments
        self.visit(node.body)
        self.visit(node.args)
        with self.push_attributes(rval_symbols=[]):
            self.visit(node.args.args)
            self.visit(node.args.vararg)
            self.visit(node.args.kwonlyargs)
            self.visit(node.args.kwarg)
            discard_set = set(self.rval_symbols)
        # throw away anything appearing in lambda body that isn't bound
        self.rval_symbols = list(set(self.rval_symbols) - discard_set)

    def visit_arg(self, node):
        self.to_add_set.append(node.arg)


def _flatten(vals):
    for v in vals:
        if isinstance(v, list):
            yield from _flatten(v)
        else:
            yield v


def _edges(lvals, rvals):
    if isinstance(lvals, list) and isinstance(rvals, list):
        yield from _edges_from_lists(lvals, rvals)
    elif isinstance(lvals, list):
        # TODO: yield edges with subscript symbols
        for left in _flatten(lvals):
            yield left, rvals
    elif isinstance(rvals, list):
        # TODO: yield edges with subscript symbols
        for right in _flatten(rvals):
            yield lvals, right
    else:
        yield lvals, rvals


def _edges_from_lists(lvals, rvals):
    if len(lvals) == len(rvals):
        for left, right in zip(lvals, rvals):
            yield from _edges(left, right)
    elif len(lvals) == 1:
        yield from _edges(lvals[0], rvals)
    elif len(rvals) == 1:
        yield from _edges(lvals, rvals[0])
    else:
        raise ValueError('Incompatible lists: %s, %s' % (lvals, rvals))


def get_assignment_lval_and_rval_symbol_refs(node: 'Union[str, ast.Assign]'):
    if isinstance(node, str):
        node = cast(ast.Assign, ast.parse(node).body[0])
    assert isinstance(node, ast.Assign)
    visitor = GetAssignmentLvalRvalSymbolRefs()
    visitor(node)
    for lval_list in visitor.lval_symbols:
        yield from _edges(lval_list, visitor.rval_symbols)



# -*- coding: utf-8 -*-
import ast
from collections import defaultdict
import logging
from typing import TYPE_CHECKING

from .attr_symbols import AttrSubSymbolChain
from .assignment_edges import get_assignment_lval_and_rval_symbol_refs
from .mixins import SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin

if TYPE_CHECKING:
    from typing import List, Set, Tuple, Union

logger = logging.getLogger(__name__)


class GetStatementLvalRvalSymbolRefs(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self):
        # TODO: current complete bipartite subgraph will add unncessary edges
        # TODO: this is actually pretty important to handle things like a.b.c properly,
        # we should switch to generating a sequence of (lvals, rvals, etc.)
        self.assignment_edges: List[Tuple[Union[str, AttrSubSymbolChain], Union[str, AttrSubSymbolChain]]] = []
        self.lval_symbol_ref_set: Set[str] = set()
        self.rval_symbol_ref_set: Set[Union[str, int]] = set()
        self.should_overwrite = True
        self.gather_rvals = True

    def __call__(self, node):
        self.visit(node)
        edges = defaultdict(set)
        if len(self.lval_symbol_ref_set) == 0 and len(self.assignment_edges) == 0:
            edges[None] = self.rval_symbol_ref_set
            return edges, self.should_overwrite
        for symbol in self.lval_symbol_ref_set:
            edges[symbol] = set(self.rval_symbol_ref_set)
        for edge in self.assignment_edges:
            # FIXME: figure out how to handle attributes in a principled manner here
            left, right = edge
            if isinstance(left, AttrSubSymbolChain):
                edges[None].add(right)
            elif isinstance(right, AttrSubSymbolChain) or right is None:
                # just get the lval in the keys
                edges[left].add(None)
                edges[left].discard(None)
            else:
                edges[left].add(right)
        return edges, self.should_overwrite

    @property
    def to_add_set(self):
        if self.gather_rvals:
            return self.rval_symbol_ref_set
        else:
            return self.lval_symbol_ref_set

    def gather_lvals_context(self):
        return self.push_attributes(gather_rvals=False)

    def gather_rvals_context(self):
        return self.push_attributes(gather_rvals=True)

    def visit_Attribute(self, node):
        # everything here is handled by the attrsub tracer
        # TODO: we'll ignore args inside of inner calls, e.g. f.g(x, y).h
        return

    def visit_Name(self, node):
        self.to_add_set.add(node.id)

    def visit_Subscript(self, node: ast.Subscript):
        # everything but the name inside the slice is handled by the attrsub tracer
        # if not self.gather_rvals and isinstance(node.slice, ast.Index) and isinstance(node.slice.value, ast.Name):
        #     with self.gather_rvals_context():
        #         self.visit(node.slice.value)
        # TODO: the ast.Name slice dependency turns out to be super hard to refresh, so until we
        #  figure out how to do it in a principled way, we'll just accept potential false negatives
        # TODO: we'll ignore args inside of inner calls, e.g. f[g](x, y)[h]
        return

    def visit_Assign(self, node):
        try:
            self.assignment_edges.extend(get_assignment_lval_and_rval_symbol_refs(node))
        except Exception as e:
            logger.warning('Exception while trying to do new-style edge computation for assignment: %s' % e)
            logger.warning('Falling back to old method...')
            with self.gather_lvals_context():
                for target in node.targets:
                    self.visit(target)
            with self.gather_rvals_context():
                self.visit(node.value)

    def visit_AnnAssign(self, node):
        with self.gather_lvals_context():
            self.visit(node.target)
        with self.gather_rvals_context():
            self.visit(node.value)

    def visit_AugAssign(self, node):
        with self.push_attributes(should_overwrite=False):
            with self.gather_lvals_context():
                self.visit(node.target)
            with self.gather_rvals_context():
                self.visit(node.value)

    def visit_Call(self, node):
        with self.gather_rvals_context():
            self.generic_visit(node)

    def visit_For(self, node):
        # skip body -- will have dummy since this visitor works line-by-line
        with self.gather_lvals_context():
            self.visit(node.target)
        with self.gather_rvals_context():
            self.visit(node.iter)

    def visit_FunctionDef(self, node):
        self.lval_symbol_ref_set.add(node.name)
        with self.gather_rvals_context():
            self.visit(node.args)

    def visit_ClassDef(self, node):
        self.lval_symbol_ref_set.add(node.name)
        with self.gather_rvals_context():
            self.visit(node.bases)
            self.visit(node.decorator_list)

    def visit_Keyword(self, node):
        self.visit(node.value)

    def visit_Starred(self, node):
        self.visit(node.value)

    def visit_AsyncFunctionDef(self, node):
        self.visit_FunctionDef(node)

    def visit_Lambda(self, node):
        # TODO: we could have lambdas defined nested inside of a lambda, in which case we also need to put the
        #  outer args in a discard set as well
        assert self.gather_rvals
        # remove node.arguments
        self.visit(node.body)
        self.visit(node.args)
        with self.push_attributes(rval_symbol_ref_set=set()):
            self.visit(node.args.args)
            self.visit(node.args.vararg)
            self.visit(node.args.kwonlyargs)
            self.visit(node.args.kwarg)
            discard_set = self.rval_symbol_ref_set
        # throw away anything appearing in lambda body that isn't bound
        self.rval_symbol_ref_set -= discard_set

    def visit_With(self, node):
        # skip body
        self.visit(node.items)

    def visit_withitem(self, node):
        with self.gather_lvals_context():
            self.visit(node.optional_vars)
        with self.gather_rvals_context():
            self.visit(node.context_expr)

    def visit_arg(self, node):
        self.to_add_set.add(node.arg)


def get_statement_symbol_edges(node: ast.AST):
    return GetStatementLvalRvalSymbolRefs()(node)



# -*- coding: utf-8 -*-
import inspect
import itertools
from typing import TYPE_CHECKING
import weakref

from IPython import get_ipython
try:
    import pandas
except ImportError:
    pandas = None

from .analysis import AttrSubSymbolChain, CallPoint
from .data_symbol import DataSymbol, DataSymbolType
from .ipython_utils import cell_counter

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union
    from .safety import DependencySafety


class Scope(object):
    GLOBAL_SCOPE_NAME = '<module>'

    def __init__(
            self,
            safety: 'DependencySafety',
            scope_name: str = GLOBAL_SCOPE_NAME,
            parent_scope: 'Optional[Scope]' = None,
    ):
        self.safety = safety
        self.scope_name = scope_name
        self.parent_scope = parent_scope  # None iff this is the global scope
        self._data_symbol_by_name: Dict[Union[str, int], DataSymbol] = {}

    def __hash__(self):
        return hash(self.full_path)

    def __str__(self):
        return str(self.full_path)

    def data_symbol_by_name(self, is_subscript=False):
        if is_subscript:
            raise ValueError('Only namespace scopes carry subscripts')
        return self._data_symbol_by_name

    @property
    def is_namespace_scope(self):
        return isinstance(self, NamespaceScope)

    @property
    def non_namespace_parent_scope(self):
        # a scope nested inside of a namespace scope does not have access
        # to unqualified members of the namespace scope
        if self.is_global:
            return None
        if self.parent_scope.is_namespace_scope:
            return self.parent_scope.non_namespace_parent_scope
        return self.parent_scope

    def make_child_scope(self, scope_name, obj_id=None):
        if obj_id is None:
            return Scope(self.safety, scope_name, parent_scope=self)
        else:
            return NamespaceScope(obj_id, self.safety, scope_name, parent_scope=self)

    def put(self, name: 'Union[str, int]', val: DataSymbol):
        self._data_symbol_by_name[name] = val
        val.containing_scope = self

    def lookup_data_symbol_by_name_this_indentation(self, name) -> 'Optional[DataSymbol]':
        return self._data_symbol_by_name.get(name, None)

    def all_data_symbols_this_indentation(self):
        return self._data_symbol_by_name.values()

    def lookup_data_symbol_by_name(self, name):
        ret = self.lookup_data_symbol_by_name_this_indentation(name)
        if ret is None and self.non_namespace_parent_scope is not None:
            ret = self.non_namespace_parent_scope.lookup_data_symbol_by_name(name)
        return ret

    @staticmethod
    def _get_name_to_obj_mapping(obj, dc) -> 'Dict[str, Any]':
        if obj is None:
            return get_ipython().ns_table['user_global']
        elif dc is not None and dc.is_subscript:
            return obj
        else:
            try:
                name_to_obj = obj.__dict__
                if (pandas is not None) and isinstance(obj, pandas.DataFrame):
                    # FIXME: hack to get it working w/ pandas, which doesn't play nicely w/ inspect.getmembers
                    name_to_obj = dict(name_to_obj)
                    name_to_obj.update(obj.to_dict())
            except:  # noqa
                return dict(inspect.getmembers(obj))
        return name_to_obj

    def get_most_specific_data_symbol_for_attrsub_chain(
            self, chain: AttrSubSymbolChain, namespaces: 'Dict[int, NamespaceScope]'
    ):
        """
        Get most specific DataSymbol for the whole chain (stops at first point it cannot find nested, e.g. a CallPoint).
        """
        cur_scope = self
        name_to_obj = get_ipython().ns_table['user_global']
        dsym = None
        obj = None
        for name in chain.symbols:
            if isinstance(name, CallPoint):
                return dsym
            dsym = cur_scope.lookup_data_symbol_by_name_this_indentation(name)
            if name_to_obj is None:
                break
            try:
                obj = Scope._get_name_to_obj_mapping(obj, dsym)[name]
            except (KeyError, IndexError, Exception):
                break
            cur_scope = namespaces.get(id(obj), None)
            if cur_scope is None:
                break
        return dsym

    def upsert_data_symbol_for_name(
            self,
            name: str,
            obj: 'Any',
            deps: 'Set[DataSymbol]',
            is_subscript,
            overwrite=True,
            is_function_def=False,
            class_scope: 'Optional[Scope]' = None,
    ):
        dc, old_dc, old_id = self._upsert_data_symbol_for_name_inner(
            name, obj, deps, is_subscript,
            overwrite=overwrite, is_function_def=is_function_def, class_scope=class_scope
        )
        # print('upsert', name, 'with deps', deps)
        self._handle_aliases(old_id, old_dc, dc)

    def _upsert_data_symbol_for_name_inner(
            self,
            name: str,
            obj: 'Any',
            deps: 'Set[DataSymbol]',
            is_subscript,
            overwrite=True,
            is_function_def=False,
            class_scope: 'Optional[Scope]' = None,
    ) -> 'Tuple[DataSymbol, Optional[DataSymbol], Optional[int]]':
        assert not (class_scope is not None and is_function_def)
        symbol_type = DataSymbolType.DEFAULT
        if is_function_def:
            assert overwrite
            assert not is_subscript
            symbol_type = DataSymbolType.FUNCTION
            # return self._upsert_function_data_symbol_for_name(name, obj, deps)
        elif class_scope is not None:
            assert overwrite
            assert not is_subscript
            symbol_type = DataSymbolType.CLASS
            # return self._upsert_class_data_symbol_for_name(name, obj, deps, class_scope)
        elif is_subscript:
            symbol_type = DataSymbolType.SUBSCRIPT
        old_id = None
        old_dc = self.lookup_data_symbol_by_name_this_indentation(name)
        if old_dc is not None and self.is_globally_accessible:
            old_id = old_dc.cached_obj_id
            # TODO: garbage collect old names (EDIT: does this happen automatically thanks to the handle_aliases logic?)
            # TODO: handle case where new dc is of different type
            if name in self.data_symbol_by_name(old_dc.is_subscript):
                old_dc.update_obj_ref(obj)
                old_dc.update_type(symbol_type)
                old_dc.update_deps(deps, overwrite=overwrite)
                return old_dc, old_dc, old_id
            else:
                # In this case, we are copying from a class and we need the dsym from which we are copying
                # as able to propagate to the new dsym.
                # Example:
                # class Foo:
                #     shared = 99
                # foo = Foo()
                # foo.shared = 42  # old_dc refers to Foo.shared here
                # Earlier, we were explicitly adding Foo.shared as a dependency of foo.shared as follows:
                # deps.add(old_dc)
                # But it turns out not to be necessary because foo depends on Foo, and changing Foo.shared will
                # propagate up the namespace hierarchy to Foo, which propagates to foo, which then propagates to
                # all of foo's namespace children (e.g. foo.shared).
                # This raises the question of whether we should draw the foo <-> Foo edge, since irrelevant namespace
                # children could then also be affected (e.g. some instance variable foo.x).
                # Perhaps a better strategy is to prevent propagation along this edge unless class Foo is redeclared.
                # If we do this, then we should go back to explicitly adding the dep as follows:
                # EDIT: added check to avoid propagating along class -> instance edge when class not redefined, so now
                # it is important to explicitly add this dep.
                deps.add(old_dc)
        dc = DataSymbol(name, symbol_type, obj, self, self.safety, parents=deps)
        self.put(name, dc)
        dc.update_deps(deps, overwrite=True)
        return dc, old_dc, old_id

    def _handle_aliases(
            self,
            old_id: 'Optional[int]',
            old_dc: 'Optional[DataSymbol]',
            dc: 'DataSymbol',
    ):
        if old_id == dc.obj_id and old_dc is dc:
            return
        if old_id is not None and old_dc is not None:
            old_alias_dcs = self.safety.aliases[old_id]
            old_alias_dcs.discard(old_dc)
            if len(old_alias_dcs) == 0:
                del self.safety.aliases[old_id]
        self.safety.aliases[dc.obj_id].add(dc)

    @property
    def is_global(self):
        return self.parent_scope is None

    @property
    def is_globally_accessible(self):
        return self.is_global or (self.is_namespace_scope and self.parent_scope.is_globally_accessible)

    @property
    def global_scope(self):
        if self.is_global:
            return self
        return self.parent_scope.global_scope

    @property
    def full_path(self) -> 'Tuple[str, ...]':
        path = (self.scope_name,)
        if self.is_global:
            return path
        else:
            return self.parent_scope.full_path + path

    @property
    def full_namespace_path(self) -> str:
        if not self.is_namespace_scope:
            return ''
        if self.parent_scope is not None:
            prefix = self.parent_scope.full_namespace_path
        else:
            prefix = ''
        if prefix:
            return f'{prefix}.{self.scope_name}'
        else:
            return self.scope_name

    def make_namespace_qualified_name(self, dc: 'DataSymbol'):
        return dc.name


class NamespaceScope(Scope):
    # TODO: support (multiple) inheritance by allowing
    #  NamespaceScopes from classes to clone their parent class's NamespaceScopes
    def __init__(self, obj: 'Any', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cloned_from: Optional[NamespaceScope] = None
        self.child_clones: List[NamespaceScope] = []
        tombstone, obj_ref, obj_id = self._update_obj_ref_inner(obj)
        self._tombstone = tombstone
        self._obj_ref = obj_ref
        self.obj_id = obj_id
        self.max_defined_timestamp = 0
        self._subscript_data_symbol_by_name: Dict[Union[int, str], DataSymbol] = {}

    def update_obj_ref(self, obj):
        tombstone, obj_ref, obj_id = self._update_obj_ref_inner(obj)
        self._tombstone = tombstone
        self._obj_ref = obj_ref
        self.obj_id = obj_id

    def _update_obj_ref_inner(self, obj):
        tombstone = False
        try:
            obj_ref = weakref.ref(obj, self._obj_reference_expired_callback)
        except TypeError:
            obj_ref = None
        obj_id = id(obj)
        return tombstone, obj_ref, obj_id

    def _obj_reference_expired_callback(self, *_):
        self._tombstone = True
        self.safety.garbage_namespace_obj_ids.add(self.obj_id)

    def data_symbol_by_name(self, is_subscript=False):
        if is_subscript:
            return self._subscript_data_symbol_by_name
        else:
            return self._data_symbol_by_name

    def clone(self, obj: 'Any'):
        cloned = NamespaceScope(obj, self.safety)
        cloned.__dict__ = dict(self.__dict__)
        cloned.cloned_from = self
        cloned.update_obj_ref(obj)
        cloned._data_symbol_by_name = {}
        cloned._subscript_data_symbol_by_name = {}
        self.child_clones.append(cloned)
        return cloned

    def shallow_clone(self, obj: 'Any'):
        cloned = NamespaceScope(obj, self.safety)
        cloned.scope_name = self.scope_name
        cloned.parent_scope = self.parent_scope
        return cloned

    def make_namespace_qualified_name(self, dc: 'DataSymbol'):
        path = self.full_namespace_path
        if path:
            if dc.is_subscript:
                return f'{path}[{dc.name}]'
            else:
                return f'{path}.{dc.name}'
        else:
            return dc.name

    def lookup_data_symbol_by_name_this_indentation(self, name, is_subscript=None):
        # TODO: specify in arguments whether `name` refers to a subscript
        if is_subscript is None:
            ret = self._data_symbol_by_name.get(name, None)
            if ret is None:
                ret = self._subscript_data_symbol_by_name.get(name, None)
        elif is_subscript:
            ret = self._subscript_data_symbol_by_name.get(name, None)
        else:
            ret = self._data_symbol_by_name.get(name, None)
        if ret is None and self.cloned_from is not None:
            ret = self.cloned_from.lookup_data_symbol_by_name_this_indentation(name)
        return ret

    def all_data_symbols_this_indentation(self, exclude_class=False, is_subscript=None) -> 'Iterable[DataSymbol]':
        if is_subscript is None:
            dsym_collections_to_chain: List[Iterable] = [
                self._data_symbol_by_name.values(), self._subscript_data_symbol_by_name.values()
            ]
        elif is_subscript:
            dsym_collections_to_chain = [self._subscript_data_symbol_by_name.values()]
        else:
            dsym_collections_to_chain = [self._data_symbol_by_name.values()]
        if self.cloned_from is not None and not exclude_class:
            dsym_collections_to_chain.append(self.cloned_from.all_data_symbols_this_indentation())
        return itertools.chain(*dsym_collections_to_chain)

    def put(self, name: 'Union[str, int]', val: DataSymbol):
        if val.is_subscript:
            self._subscript_data_symbol_by_name[name] = val
        else:
            self._data_symbol_by_name[name] = val
        val.containing_scope = self

    def refresh(self):
        self.max_defined_timestamp = cell_counter()

    def get_earliest_ancestor_containing(self, obj_id: int, is_subscript: bool) -> 'Optional[NamespaceScope]':
        # TODO: test this properly
        ret = None
        if self.namespace_parent_scope is not None:
            ret = self.namespace_parent_scope.get_earliest_ancestor_containing(obj_id, is_subscript)
        if ret is not None:
            return ret
        set_to_check = map(lambda dsym: dsym.obj_id, self.all_data_symbols_this_indentation(is_subscript=is_subscript))
        if obj_id in set_to_check:
            return self
        else:
            return None

    @property
    def namespace_parent_scope(self) -> 'Optional[NamespaceScope]':
        if self.parent_scope is not None and isinstance(self.parent_scope, NamespaceScope):
            return self.parent_scope
        return None



# -*- coding: utf-8 -*-

from nbsafety.analysis.stmt_edges import get_statement_symbol_edges
from nbsafety.analysis.lineno_stmt_map import compute_lineno_to_stmt_mapping


def get_statement_lval_and_rval_symbols(node):
    edges, _ = get_statement_symbol_edges(node)
    lvals = set(edges.keys())
    if len(edges) == 0:
        rvals = set()
    else:
        rvals = set.union(*edges.values())
    return lvals - {None}, rvals - {None}


def test_classes():
    code = """
class Foo(object):
    pass
    
class Bar(Foo):
    pass
    
class Baz(Foo, Bar):
    pass
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'Foo'}
    assert rvals == {'object'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[4])
    assert lvals == {'Bar'}
    assert rvals == {'Foo'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[7])
    assert lvals == {'Baz'}
    assert rvals == {'Foo', 'Bar'}


def test_for_loop():
    code = """
for i in range(10):
    a = i
    b = a + i
    lst = [a, b]
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'i'}
    assert rvals == {'range'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[2])
    assert lvals == {'a'}
    assert rvals == {'i'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[3])
    assert lvals == {'b'}
    assert rvals == {'a', 'i'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[4])
    assert lvals == {'lst'}
    assert rvals == {'a', 'b'}


def test_context_manager():
    code = """
fname = 'file.txt'
with open(fname) as f:
    contents = f.read()
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[2])
    assert lvals == {'f'}
    assert rvals == {'fname', 'open'}
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[3])
    assert lvals == {'contents'}
    # attributes should be skipped
    assert rvals == set()


def test_deeply_nested_arguments():
    mapping = compute_lineno_to_stmt_mapping('x = f(f(f(f(y, z=7), z=9), z=w), 10)')
    lvals, rvals = get_statement_lval_and_rval_symbols(mapping[1])
    assert lvals == {'x'}
    assert rvals == {'f', 'y', 'w'}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import atexit
import platform
import os
from subprocess import check_call

from setuptools import setup, find_packages
import versioneer

pkg_name = 'nbsafety'


def _post_install_hook():
    python_command_suffix = ''
    if platform.system().lower().startswith('win'):
        python_command_suffix = '.exe'
    check_call(f'python{python_command_suffix} -m {pkg_name}.install'.split())


atexit.register(_post_install_hook)


def read_file(fname):
    with open(fname, 'r') as f:
        return f.read()


history = read_file('HISTORY.rst')
requirements = read_file('requirements.txt').strip().split()
setup(
    name=pkg_name,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='Stephen Macke',
    author_email='stephen.macke@gmail.com',
    description='Fearless interactivity for Jupyter notebooks.',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/nbsafety-project/nbsafety',
    packages=find_packages(exclude=['binder', 'docs', 'scratchspace', 'notebooks', 'img', 'test']),
    include_package_data=True,
    install_requires=requirements,
    license='BSD-3-Clause',
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)

# python setup.py sdist bdist_wheel --universal
# twine upload dist/*
